package ch.rakudave.jnetmap.util;

import java.awt.Graphics;
import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.Reader;
import java.io.StringWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.zip.ZipInputStream;

import javax.imageio.ImageIO;
import javax.swing.JOptionPane;

import ch.rakudave.jnetmap.model.Connection;
import ch.rakudave.jnetmap.model.device.Device;
import ch.rakudave.jnetmap.plugins.Plugin;
import ch.rakudave.jnetmap.util.logging.Logger;
import ch.rakudave.jnetmap.view.components.SplashScreen;
import edu.uci.ics.jung.visualization.VisualizationViewer;

/**
 * @author rakudave
 */
public class IO {
	private static final int bufferSize = 4096;
	public static final File userDir = new File(System.getProperty("user.home") + "/.jNetMap");
	public static final File pluginDir = new File(userDir, "/plugins");
	public static final File langDir = new File(IO.userDir, "/lang");
	public static final boolean isUnix = !System.getProperty("os.name").startsWith("Windows");
	public static final boolean isOSX = System.getProperty("os.name").startsWith("Mac OS X");

	/**
	 * Copy a file or directory recursively
	 * @param source source file
	 * @param dest destination file
	 * @param ignoreList file-names to ignore
	 * @return success
	 */
	public static boolean copy(File source, File dest, List<String> ignoreList) {
		if (ignoreList != null && ignoreList.contains(source.getName())) return true;
		boolean result = true;
		if (source.isDirectory()) {
			Logger.trace("Copying dir "+dest.getName());
			dest.mkdirs();
			String list[] = source.list();
			for (int i = 0; i < list.length; i++) {
				result &= copy(new File(source, list[i]), new File(dest, list[i]), ignoreList);
			}
		} else {
			Logger.trace("Copying file "+dest.getName());
			FileInputStream fin = null;
			FileOutputStream fout = null;
			byte[] buffer = new byte[bufferSize];
			int bytesRead;
			try {
				fin = new FileInputStream(source);
				fout = new FileOutputStream(dest);
				while ((bytesRead = fin.read(buffer)) >= 0) {
					fout.write(buffer, 0, bytesRead);
				}
				return true;
			} catch (IOException e) {
				Logger.error("Failed to open I/O-Stream(s)", e);
				return false;
			} finally {
				try {
					fin.close();
					fout.close();
				} catch (IOException e) {}
			}
		}
		return result;
	}
		
	public static boolean copy(InputStream in, OutputStream out) {
		try {
			byte[] buf = new byte[bufferSize];
			int len;
			while ((len = in.read(buf)) > 0) {
				out.write(buf, 0, len);
			}
			return true;
		} catch (IOException e) {
			Logger.error("Failed to open I/O-stream(s)", e);
		} finally {
			try {
				in.close();
				out.close();
			} catch (Exception e) {}
		}
		return false;
	}
	
	public static boolean copy(InputStream in, String destPath) {
		File dest = new File(destPath);
		FileOutputStream out = null;
		try {
			out = new FileOutputStream(dest);
			return copy(in, out);
		} catch (Exception e) {
			Logger.error("Failed to open output-stream", e);
			return false;
		} finally {
			try {out.close();} catch (Exception e) {}
		}
	}

	/**
	 * create user directory and install preferences, languages and icons
	 */
	public static void installUserFiles() {
		Logger.info("Setting up user directory...");
		if (!userDir.exists()) userDir.mkdir();
		try {
			Settings.file.createNewFile();
			File jar = new File(System.getProperty("java.class.path").split((isUnix)?":":";")[0]);
			if (jar.isDirectory()) { // not a jar, find res and copy (for IDE and non-jar execution)
				Logger.trace("Non-jar exec detected");
				copy(new File(jar.getParent()+"/res"), userDir, Arrays.asList(".svn"));
			} else { // extract relevant files from jar
				Logger.trace("jar exec detected");
				ZipInputStream zip = new ZipInputStream(new FileInputStream(jar));
				ZipEntry ze; String name;
				while ((ze = zip.getNextEntry()) != null) {
					name = ze.getName();
					Logger.trace("zipentry "+name);
					if (name.startsWith("icons") || name.startsWith("lang") || name.startsWith("img") 
							|| name.startsWith("devices") || name.startsWith("plugins") || name.startsWith("guide.pdf")
							|| name.startsWith("ports") || name.startsWith("MIB") || name.startsWith("OUI")) {
						if (ze.isDirectory()) {
							Logger.trace("mkdir "+name);
							new File(userDir.getAbsolutePath()+"/"+name).mkdir();
						} else {
							if (!(name.startsWith("plugins") && new File(IO.userDir.getAbsolutePath()+"/"+name).exists()))
								copy(IO.class.getResourceAsStream("/"+name), IO.userDir.getAbsolutePath()+"/"+name);
						}
					}
				}
				zip.close();
			}
		} catch (Exception e) {
			Logger.fatal("Failed to set up user directory!", e);
			userDir.delete();
			SplashScreen.dispose();
			JOptionPane.showMessageDialog(null, 
					"Failed to set up user directory!\n" +
					"This is fatal, exiting...", 
					"FATAL ERROR", JOptionPane.ERROR_MESSAGE);
			System.exit(1);
		}
	}
	
	public static boolean unzip(File zipFile, File destination) {
		if (zipFile == null || (!zipFile.exists() || !zipFile.canRead() 
				|| !zipFile.getName().contains(".zip"))) return false;
		ZipFile zip = null;
		try {
			boolean success = true;
			zip = new ZipFile(zipFile);
			Enumeration<? extends ZipEntry> oEnum = zip.entries();
            while(oEnum.hasMoreElements()) {
                ZipEntry zipEntry = oEnum.nextElement();
				if (zipEntry.isDirectory()) {
					success &= (new File(destination.getAbsoluteFile() + "/" + zipEntry.getName())).mkdir();
				} else {
					success &= copy(zip.getInputStream(zipEntry), destination.getAbsoluteFile() 
							+ "/" + zipEntry.getName());
				}
            }
            return success;
		} catch (IOException e) {
			Logger.error("Failed to open ZIP-file", e);
			// attempt cleanup
			(new File(destination, zipFile.getName().replace(".zip", ""))).delete();
			return false;
		} finally {
			try {zip.close();} catch (IOException e) {}
		}
	}
	
	public static boolean exportImage(VisualizationViewer<Device, Connection> in, File out) {
		System.out.println(in.getSize());
		BufferedImage buff = new BufferedImage(in.getWidth(), in.getHeight(),
				java.awt.image.BufferedImage.TYPE_INT_RGB);
		Graphics g = buff.getGraphics();
		in.setDoubleBuffered(false);
		in.paintAll(g);
		in.setDoubleBuffered(true);
		g.dispose();
		try {
			ImageIO.write(buff, "png", out);
			return true;
		} catch (IOException e) {
			Logger.error("Unable to export image", e);
			return false;
		}
	}

	public static String getString(InputStream is) {
		if (is == null) return null;
		Writer writer = new StringWriter();
		char[] buffer = new char[bufferSize];
		try {
			Reader reader = new BufferedReader(new InputStreamReader(is, "UTF-8"));
			int n;
			while ((n = reader.read(buffer)) != -1) {
				writer.write(buffer, 0, n);
			}
		} catch (Exception e) {
			Logger.error("Unable to read string", e);
			return null;
		} finally {
			try {is.close();} catch (IOException e) {}
		}
		return writer.toString();
	}

	public static boolean tryRemovePlugin(Plugin p) {
		// since jspf does not know where a plugin came from, lets do this the hard way
		// (aka. open every jar an look for matching class names)
		for (File f : new File(IO.userDir, "plugins").listFiles()) {
			if (f.getName().contains(".jar")) {
				ZipFile zip = null;
				try {
					zip = new ZipFile(f);
					Enumeration<? extends ZipEntry> oEnum = zip.entries();
		            while(oEnum.hasMoreElements()) {
		                ZipEntry zipEntry = oEnum.nextElement();
						if (!zipEntry.isDirectory() && zipEntry.getName().contains(
								p.getClass().getSimpleName()+".class")) {
							return f.delete();
						}
		            }
				} catch (Exception ex) {
					Logger.warn("Unable to read jar-file "+f, ex);
				} finally {
					try {zip.close();} catch (IOException e) {}
				}
			}
		}
		return false;
	}
	

	/**
	 * @author Jan Goyvaerts
	 */
	public static String[] splitCommandArgs(String command, String args) {
		List<String> matchList = new ArrayList<String>();
			matchList.add(command);
		Pattern regex = Pattern.compile("[^\\s\"']+|\"([^\"]*)\"|'([^']*)'");
		Matcher regexMatcher = regex.matcher(args);
		while (regexMatcher.find()) {
		    if (regexMatcher.group(1) != null) {
		        // Add double-quoted string without the quotes
		        matchList.add(regexMatcher.group(1));
		    } else if (regexMatcher.group(2) != null) {
		        // Add single-quoted string without the quotes
		        matchList.add(regexMatcher.group(2));
		    } else {
		        // Add unquoted word
		        matchList.add(regexMatcher.group());
		    }
		}
		String[] out = new String[matchList.size()];
		for (int i = 0; i < out.length; i++) {
			out[i] = matchList.get(i);
		}
		return out;
	}
}